{% extends "epic/base_layout.html" %}
{% load crispy_forms_tags %}

{% block content %}
{{ block.super }}

<form method="POST">
  <div class="form-horizontal">
    {% crispy form %}
  </div>

  <div class="form-actions">
    <input type="submit" name="save" value="Save" class="btn btn-primary">
    <input type="submit" name="cancel" value="Cancel" class="btn btn-default">
  </div>
</form>

{% endblock content %}
