{% extends "epic/base_layout.html" %}
{% load crispy_forms_tags %}

{% block extrahead %}
{{ block.super }}
{% include "autocomplete_light/static.html" %}
{{ order_form.media }}	<!-- for django-bootstrap3-datetimepicker -->
<link rel="stylesheet" type="text/css"
      href="{{ STATIC_URL }}epic/css/epic.css" />
<script src="{{ STATIC_URL }}epic/js/editing.js"></script>
{% endblock extrahead %}

{% block content %}
{{ block.super }}

<form method="POST">

<div class="form-horizontal">
  {% crispy order_form %}
</div>

{% crispy item_form %}

<div class="form-actions">
  <input type="submit" name="save" value="Save" class="btn btn-primary">
  <input type="submit" name="save-and-done" value="Save & Done"
	 class="btn btn-default">
  <input type="submit" name="cancel" value="Cancel"
	 class="btn btn-default">
</div>
</form>

<br>

{% endblock content %}
