.epic_deleted {
    text-decoration: line-through;
    color: red;
}

.epic_ok {
    color: green;
}

.epic_alert {
    color: red;
}

.epic_excess {
    color: green;
}

.epic_shortage {
    color: red;
}

.epic_substitute_deck {
    height: inherit;
}

.epic_substitute_element, .epic_substitute_add {
    border: 1px solid #ccc;
    border-radius: 4px;
    padding: 6px;
}

.epic_substitute_add input {
    border: 1px solid #ccc;
    border-radius: 4px;
}
